from collections import defaultdict
from collections.abc import Sequence
from dataclasses import dataclass, field
from enum import Enum
from typing import Dict

from termcolor import colored


class Result(Enum):
    PASS = colored("PASS", "green")
    WARN = colored("WARN", "yellow")
    FAIL = colored("FAIL", "red")
    SKIP = colored("SKIP", "blue")


@dataclass(repr=False)
class Message:
    result: Result
    check_name: str
    message: str
    result_code: str = "ok"
    context: dict = field(default_factory=dict)
    fixes: list = field(default_factory=list)

    def __repr__(self) -> str:
        return self.result.value + ": " + self.message

    def __hash__(self) -> int:
        hsh = hash(self.check_name + self.message + self.result.value)
        return hsh

    def __eq__(self, other) -> bool:
        return self.check_name == other.check_name and self.message == other.message


class Reporter(Sequence):
    def __init__(self) -> None:
        self.results = {}

    def __getitem__(self, index):
        return list(self.results.keys())[index]

    def __len__(self) -> int:
        return len(self.results.keys())

    def okay(self, **kwargs) -> None:
        self.results[Message(result=Result.PASS, **kwargs)] = 1

    def warn(self, **kwargs) -> None:
        self.results[Message(result=Result.WARN, **kwargs)] = 1

    def fail(self, **kwargs) -> None:
        self.results[Message(result=Result.FAIL, **kwargs)] = 1

    def skip(self, **kwargs) -> None:
        self.results[Message(result=Result.SKIP, **kwargs)] = 1

    @property
    def is_unknown(self) -> bool:
        return not self.passes and not self.fails

    @property
    def is_success(self) -> bool:
        return len(self.passes) > 0 and not self.fails

    def is_nearly_success(self, nearly: int) -> bool:
        return not self.is_success and self.count_fixes <= nearly

    @property
    def passes(self) -> list:
        return [x for x in self.results if x.result == Result.PASS]

    @property
    def fails(self) -> list:
        return [x for x in self.results if x.result == Result.FAIL]

    @property
    def warns(self) -> list:
        return [x for x in self.results if x.result == Result.WARN]

    def unique_fixes(self) -> Dict[str, list[str]]:
        fixes = defaultdict(list)
        for result in self.results:
            for fix in result.fixes:
                fixtype = fix["type"]
                fixthing = fix["thing"]
                if fixthing not in fixes[fixtype]:
                    fixes[fixtype].append(fixthing)
        return fixes

    @property
    def count_fixes(self) -> int:
        """Return the number of fixes required to pass all checks"""
        fixes = self.unique_fixes()
        return sum(len(stuff) for stuff in fixes.values())



import unicodedata
from typing import Iterator, List

from youseedee import ucd_data

from shaperglot.checks import NoOrphanedMarksCheck, ShapingDiffersCheck
from shaperglot.checks.common import ShaperglotCheck

ZWJ = "\u200D"
MANUALLY_DEFINED_MARKS = {
    'ar_Arab': [
        ["\u064E\u0651", 'FATHA+SHADDA'],
        ["\u064B\u0651", 'FATHATAN+SHADDA'],
        ["\u0650\u0651", 'KASRA+SHADDA'],
        ["\u064D\u0651", 'KASRATAN+SHADDA'],
        ["\u064F\u0651", 'DAMMA+SHADDA'],
        ["\u064C\u0651", 'DAMMATAN+SHADDA'],
    ]
}


def get_joining_type(char: str) -> str:
    return ucd_data(ord(char)).get("Joining_Type", "U")


class ArabicProvider:
    @classmethod
    def fill(cls, language) -> None:
        if language["script"] != "Arab":
            return
        if "exemplarChars" not in language or "base" not in language["exemplarChars"]:
            return
        # Assemble list of marks
        check_these_marks = []
        for character in language["exemplarChars"].get("marks", []):
            if unicodedata.category(character) == "Mn":
                check_these_marks.append([character, unicodedata.name(character)])
        if language["id"] in MANUALLY_DEFINED_MARKS:
            check_these_marks.extend(MANUALLY_DEFINED_MARKS[language["id"]])
        for character in language["exemplarChars"]["base"]:
            if unicodedata.category(character).startswith("L"):
                joining_type = get_joining_type(character)
                if joining_type in ("R", "D"):
                    language["shaperglot_checks"].extend(
                        cls.shaping_checks(ZWJ, character, '', 'fina')
                    )
                    language["shaperglot_checks"].extend(
                        cls.mark_checks(ZWJ, character, '', check_these_marks, 'fina')
                    )
                if joining_type == "D":
                    language["shaperglot_checks"].extend(
                        cls.shaping_checks(ZWJ, character, ZWJ, 'medi')
                    )
                    language["shaperglot_checks"].extend(
                        cls.mark_checks(ZWJ, character, ZWJ, check_these_marks, 'medi')
                    )
                    language["shaperglot_checks"].extend(
                        cls.shaping_checks('', character, ZWJ, 'init')
                    )
                    language["shaperglot_checks"].extend(
                        cls.mark_checks('', character, ZWJ, check_these_marks, 'init')
                    )
                language["shaperglot_checks"].extend(
                    cls.mark_checks('', character, '', check_these_marks, 'isol')
                )

    @classmethod
    def mark_checks(  # pylint: disable=too-many-arguments
        cls,
        pre_context: str,
        character: str,
        post_context: str,
        check_these_marks: List[str],
        position: str,
    ) -> Iterator[ShaperglotCheck]:
        for marks, rationale in check_these_marks:
            cname = unicodedata.name(character)
            yield NoOrphanedMarksCheck(
                {
                    'input': {"text": pre_context + character + marks + post_context},
                    'rationale': f"{rationale} on top of {cname} (.{position})",
                }
            )

    @classmethod
    def shaping_checks(
        cls, pre_context: str, character: str, post_context: str, position: str
    ) -> Iterator[ShaperglotCheck]:
        yield ShapingDiffersCheck(
            {
                'inputs': [
                    {
                        'text': pre_context + character + post_context,
                        'features': {
                            "init": False,
                            "medi": False,
                            "fina": False,
                        },
                    },
                    {'text': pre_context + character + post_context},
                ],
                'rationale': f".{position} version of {unicodedata.name(character)}",
            }
        )



import os
import yaml
import sys
import unicodedata
from gflanguages import LoadLanguages, LoadScripts
from google.protobuf.json_format import MessageToDict

gflangs = LoadLanguages()
gfscripts = LoadScripts()

ZWJ = "\u200D"
MANUALLY_DEFINED_MARKS = {
    'ar_Arab': [
        ["\u064E\u0651", 'FATHA+SHADDA'],
        ["\u064B\u0651", 'FATHATAN+SHADDA'],
        ["\u0650\u0651", 'KASRA+SHADDA'],
        ["\u064D\u0651", 'KASRATAN+SHADDA'],
        ["\u064F\u0651", 'DAMMA+SHADDA'],
        ["\u064C\u0651", 'DAMMATAN+SHADDA'],
    ]
}

with open('./data/ArabicShaping.txt', 'r') as f:
    arabic_shaping = [
        line
        for line in f.read().splitlines()
        if not line.startswith('#') and line.strip()
    ]


def create_file(profile_name):
    path = "./shaperglot/languages/"
    os.makedirs(os.path.dirname(path), exist_ok=True)
    profile = open(path + profile_name, "w")
    return profile


def build_results(item, new_profile):
    profile_name = '%s.yaml' % item
    profile = create_file(profile_name)
    profile.write("#auto-generated using sg-bulk-builder-arabic.py\n")
    yaml.safe_dump(new_profile, profile, allow_unicode=True, sort_keys=False)
    print("Building " + item + ".yaml")


def get_joining_type(char):
    hex_code = str(hex(ord(char))).split("x")[1].zfill(4).upper()
    for line in arabic_shaping:
        if line.startswith(f"{hex_code};"):
            return line.split(';')[2].strip()


def mark_checks(pre_context, character, post_context, check_these_marks, position):
    for marks, rationale in check_these_marks:
        shaping_test = {
            'check': 'no_orphaned_marks',
            'input': {"text": pre_context + character + marks + post_context},
            'rationale': f"{rationale} on top of {unicodedata.name(character)} (.{position})",
        }
        yield shaping_test


def shaping_checks(pre_context, character, post_context, position):
    shaping_test = {
        'check': 'shaping_differs',
        'inputs': [
            {
                'text': pre_context + character + post_context,
                'features': {
                    "init": False,
                    "medi": False,
                    "fina": False,
                },
            },
            {'text': pre_context + character + post_context},
        ],
        'rationale': f".{position} version of {unicodedata.name(character)}",
    }
    yield shaping_test


def main():
    for language in gflangs:
        tests = []
        if gflangs[language].script == "Arab" and gflangs[language].exemplar_chars.base:
            # Assemble list of marks
            check_these_marks = []
            for character in gflangs[language].exemplar_chars.marks:
                if unicodedata.category(character) == "Mn":
                    check_these_marks.append([character, unicodedata.name(character)])
            if language in MANUALLY_DEFINED_MARKS:
                check_these_marks.extend(MANUALLY_DEFINED_MARKS[language])

            for character in gflangs[language].exemplar_chars.base:
                if unicodedata.category(character).startswith("L"):
                    joining_type = get_joining_type(character)
                    if joining_type in ("R", "D"):
                        # .fina
                        tests.extend(shaping_checks(ZWJ, character, '', 'fina'))
                        tests.extend(
                            mark_checks(ZWJ, character, '', check_these_marks, 'fina')
                        )
                    if joining_type == "D":
                        # .medi
                        tests.extend(shaping_checks(ZWJ, character, ZWJ, 'medi'))
                        tests.extend(
                            mark_checks(ZWJ, character, ZWJ, check_these_marks, 'medi')
                        )
                        # .init
                        tests.extend(shaping_checks('', character, ZWJ, 'init'))
                        tests.extend(
                            mark_checks('', character, ZWJ, check_these_marks, 'init')
                        )
                    # .isol
                    tests.extend(
                        mark_checks('', character, '', check_these_marks, 'isol')
                    )

        if tests:
            build_results(language, tests)


if __name__ == '__main__':
    main()



# pylint: disable=C0302
import unicodedata
from typing import Iterator, List

from shaperglot.checks.common import ShaperglotCheck
from shaperglot.checks.no_orphaned_marks import NoOrphanedMarksCheck
from shaperglot.checks.shaping_differs import ShapingDiffersCheck
from shaperglot.checks.unencoded_variants import UnencodedVariantsCheck

# fmt: off
isoconv = {
    "aar": "aa", "afr": "af", "aka": "ak", "amh": "am", "bam": "bm",
    "ewe": "ee", "ful": "ff", "hau": "ha", "her": "hz", "ibo": "ig",
    "kik": "ki", "kin": "rw", "kon": "kg", "kua": "kj", "lin": "ln",
    "lub": "lu", "lug": "lg", "mlg": "mg", "nbl": "nr", "nde": "nd",
    "ndo": "ng", "nya": "ny", "orm": "om", "run": "rn", "sag": "sg",
    "sna": "sn", "som": "so", "sot": "st", "ssw": "ss", "swa": "sw",
    "tir": "ti", "tsn": "tn", "tso": "ts", "twi": "tw", "ven": "ve",
    "wol": "wo", "xho": "xh", "yor": "yo",
}

afr_tags = [
    "aaa", "aab", "aal", "aao", "aar", "aas", "aba", "abb", "abi", "abm",
    "abn", "abo", "abr", "abu", "acb", "acd", "ach", "acp", "acq", "acw",
    "acx", "acz", "ada", "add", "ade", "adh", "adj", "adq", "adu", "aeb",
    "aec", "ael", "afb", "afe", "afn", "afo", "afr", "aft", "afu", "agb",
    "agc", "agh", "agj", "agq", "ags", "aha", "ahg", "ahi", "ahl", "ahm",
    "ahn", "ahp", "ahs", "aik", "aiw", "aiy", "aja", "ajg", "aju", "ajw",
    "aka", "akd", "akf", "akp", "aks", "aku", "akw", "ala", "ald", "alf",
    "alw", "alz", "amb", "amf", "amh", "amj", "amo", "anc", "anf", "ank",
    "ann", "anu", "anv", "anw", "any", "apd", "aqd", "aqg", "arb", "arq",
    "arv", "ary", "arz", "asa", "asg", "asj", "ass", "asv", "atg", "ati",
    "ato", "atu", "aug", "auh", "auj", "aum", "avi", "avn", "avu", "awc",
    "awn", "awo", "axk", "ayb", "aye", "ayg", "ayi", "ayk", "ayl", "ayu",
    "azo", "bab", "baf", "bag", "bam", "bas", "bau", "bav", "baw", "bax",
    "bba", "bbe", "bbg", "bbi", "bbj", "bbk", "bbm", "bbo", "bbp", "bbq",
    "bbs", "bbt", "bbu", "bbw", "bbx", "bby", "bcb", "bce", "bcg", "bci",
    "bcn", "bcp", "bcq", "bcs", "bct", "bcv", "bcw", "bcy", "bcz", "bda",
    "bde", "bdh", "bdi", "bdj", "bdm", "bdn", "bdo", "bdp", "bds", "bdt",
    "bdu", "beb", "bec", "beh", "bej", "bem", "beq", "bes", "bet", "bev",
    "bex", "bez", "bfa", "bfd", "bff", "bfj", "bfl", "bfm", "bfo", "bfp",
    "bga", "bgf", "bgj", "bgo", "bgu", "bhs", "bhy", "bib", "bid", "bif",
    "bij", "bil", "bim", "bin", "bip", "biv", "biw", "biz", "bja", "bjg",
    "bji", "bjo", "bjt", "bju", "bjv", "bjw", "bka", "bkc", "bkf", "bkg",
    "bkh", "bkj", "bkm", "bko", "bkp", "bkt", "bkv", "bkw", "bky", "ble",
    "blh", "bli", "blm", "blo", "blv", "bly", "bma", "bmb", "bmd", "bme",
    "bmf", "bmg", "bmi", "bml", "bmo", "bmq", "bms", "bmv", "bmw", "bng",
    "bni", "bnl", "bnm", "bnx", "bnz", "bob", "boe", "bof", "boh", "bok",
    "bol", "bom", "boo", "bot", "bou", "bov", "box", "boy", "boz", "bpd",
    "bpj", "bqa", "bqc", "bqd", "bqf", "bqg", "bqj", "bqk", "bqm", "bqo",
    "bqp", "bqt", "bqu", "bqv", "bqw", "bqx", "bqz", "brf", "bri", "brk",
    "brl", "brm", "brt", "bsc", "bse", "bsf", "bsi", "bsj", "bsl", "bso",
    "bsp", "bsq", "bsr", "bss", "bst", "bsv", "bsw", "bsx", "bta", "btc",
    "bte", "btf", "btg", "btt", "btu", "bub", "bud", "buf", "bui", "buj",
    "bum", "bun", "bus", "buu", "buw", "bux", "buy", "buz", "bva", "bvb",
    "bvf", "bvg", "bvh", "bvi", "bvj", "bvm", "bvo", "bvq", "bvw", "bvx",
    "bwc", "bwg", "bwh", "bwj", "bwl", "bwo", "bwq", "bwr", "bws", "bwt",
    "bwu", "bww", "bwy", "bwz", "bxb", "bxc", "bxe", "bxg", "bxk", "bxl",
    "bxp", "bxq", "bxs", "bxv", "bxw", "byb", "byc", "byf", "byg", "byi",
    "byj", "byn", "byp", "bys", "byt", "byv", "bza", "bze", "bzm", "bzo",
    "bzv", "bzw", "bzx", "bzy", "bzz", "cae", "cbj", "cbo", "cbq", "cce",
    "ccg", "cch", "ccj", "cdr", "cen", "cet", "cfa", "cfd", "cfg", "cgg",
    "chw", "cib", "cie", "cjk", "ckl", "cko", "ckq", "ckx", "cky", "cla",
    "cli", "cll", "cme", "cmt", "cnu", "coh", "cop", "cou", "cpn", "cpo",
    "cra", "cry", "csk", "cug", "cuh", "cuv", "cwa", "cwb", "cwe", "cwt",
    "daa", "dae", "dag", "dai", "daj", "dal", "dam", "das", "dau", "dav",
    "dba", "dbb", "dbd", "dbg", "dbi", "dbm", "dbo", "dbp", "dbq", "dbr",
    "dbt", "dbu", "dbv", "dbw", "ddd", "dde", "ddn", "dds", "dec", "dee",
    "deg", "dek", "deq", "dez", "dga", "dgb", "dgd", "dgh", "dgi", "dgk",
    "dgl", "dgs", "dhm", "dhs", "dib", "dic", "did", "dig", "dii", "dik",
    "dil", "dim", "dio", "dip", "dir", "diu", "diw", "diz", "djc", "dje",
    "djm", "dks", "dkx", "dlk", "dma", "dmb", "dme", "dmm", "dmo", "dmx",
    "dne", "dnj", "dnn", "dno", "doe", "doh", "doo", "dop", "dos", "dot",
    "dov", "dow", "dox", "doy", "doz", "drb", "dri", "drs", "dsh", "dsi",
    "dsq", "dti", "dtk", "dtm", "dtn", "dto", "dts", "dtt", "dtu", "dua",
    "dug", "dur", "dux", "duz", "dwa", "dwr", "dya", "dyi", "dym", "dyo",
    "dyu", "dza", "dzg", "dzn", "ebg", "ebo", "ebr", "ebu", "efa", "efe",
    "efi", "ega", "ego", "ehu", "eja", "eka", "eke", "eki", "ekm", "eko",
    "ekp", "ekr", "elh", "eli", "elm", "elo", "ema", "emk", "emn", "enb",
    "enn", "env", "enw", "eot", "epi", "erh", "etb", "eto", "ets", "etu",
    "etx", "evh", "ewe", "ewo", "eyo", "eza", "eze", "fah", "fak", "fal",
    "fam", "fan", "fap", "fer", "ffm", "fgr", "fia", "fie", "fip", "fir",
    "fkk", "fli", "fll", "flr", "fmp", "fni", "fod", "fom", "fon", "fub",
    "fuc", "fue", "fuf", "fuh", "fui", "fuj", "fum", "fuq", "fuu", "fuv",
    "fvr", "fwe", "gaa", "gab", "gax", "gaz", "gbg", "gbh", "gbn", "gbo",
    "gbp", "gbq", "gbr", "gbs", "gbv", "gbx", "gby", "gde", "gdf", "gdi",
    "gdk", "gdl", "gdm", "gdu", "gea", "gec", "ged", "geg", "gej", "gek",
    "gel", "geq", "gev", "gew", "gex", "gey", "gez", "ggb", "ggu", "gha",
    "ghl", "gho", "gic", "gid", "gie", "gis", "gix", "giz", "gji", "gjn",
    "gke", "gkn", "gkp", "gku", "glc", "glj", "glo", "glr", "glu", "glw",
    "gly", "gmd", "gmm", "gmn", "gmv", "gmx", "gmz", "gna", "gnd", "gne",
    "gng", "gnh", "gnj", "gnk", "gnz", "goa", "god", "gof", "gog", "gol",
    "gou", "gow", "gox", "goy", "gpa", "gqa", "gqr", "grd", "grh", "grj",
    "grr", "gru", "grv", "gry", "gsl", "gso", "gua", "gud", "guk", "gur",
    "guw", "gux", "guz", "gvl", "gvm", "gwa", "gwb", "gwd", "gwe", "gwg",
    "gwj", "gwn", "gwr", "gwx", "gxx", "gya", "gye", "gyg", "gyi", "gyl",
    "gza", "hae", "hag", "han", "haq", "har", "hau", "hav", "hay", "hba",
    "hbb", "hbn", "hdy", "hed", "heh", "hem", "her", "hgm", "hhr", "hia",
    "hig", "hij", "hio", "hka", "hke", "hmb", "hna", "hnh", "hod", "hoe",
    "hol", "hom", "hoo", "hor", "hoz", "hts", "huc", "hum", "hwa", "hwo",
    "hya", "ibb", "ibe", "ibm", "ibn", "ibo", "ibr", "iby", "ica", "ich",
    "ida", "idc", "idd", "ide", "idr", "ids", "idu", "ife", "ifm", "igb",
    "ige", "igl", "igw", "ihi", "ijc", "ije", "ijj", "ijn", "ijs", "iki",
    "ikk", "ikl", "iko", "ikp", "ikv", "ikw", "ikx", "ikz", "ilb", "ilv",
    "ior", "iqw", "iri", "irk", "ish", "isi", "isn", "iso", "isu", "itm",
    "its", "itw", "iya", "iyo", "iyx", "izr", "izz", "jab", "jad", "jaf",
    "jbe", "jbn", "jbu", "jek", "jen", "jer", "jeu", "jgb", "jgk", "jgo",
    "jia", "jib", "jid", "jie", "jii", "jim", "jit", "jjr", "jku", "jle",
    "jmb", "jmc", "jmi", "jmr", "jms", "jni", "jnj", "job", "jod", "jow",
    "jrr", "jrt", "jub", "jud", "juh", "juk", "jum", "juo", "juu", "juw",
    "jwi", "jyy", "kab", "kad", "kah", "kai", "kaj", "kam", "kao", "kbj",
    "kbl", "kbn", "kbo", "kbp", "kbr", "kbs", "kby", "kbz", "kcc", "kce",
    "kcf", "kcg", "kch", "kci", "kcj", "kck", "kcm", "kcp", "kcq", "kcr",
    "kcs", "kcu", "kcv", "kcw", "kcx", "kcy", "kcz", "kdc", "kde", "kdg",
    "kdh", "kdi", "kdj", "kdl", "kdm", "kdn", "kdp", "kdu", "kdx", "kdz",
    "keb", "kec", "ked", "kef", "keg", "kel", "ken", "keo", "ker", "kes",
    "keu", "kez", "kfl", "kfn", "kfo", "kfz", "kga", "kgo", "kgt", "khj",
    "khq", "khu", "khx", "khy", "kia", "kib", "kid", "kie", "kik", "kil",
    "kin", "kiv", "kiz", "kka", "kkd", "kke", "kki", "kkj", "kkm", "kko",
    "kkq", "kkr", "kks", "kku", "kkw", "klc", "klf", "klk", "klo", "klu",
    "kma", "kmb", "kme", "kmi", "kmp", "kmq", "kmw", "kmy", "kna", "knc",
    "knf", "kng", "kni", "knk", "kno", "knp", "knu", "knw", "kny", "knz",
    "koc", "koe", "kof", "koh", "kon", "koo", "koq", "kot", "kou", "kov",
    "kow", "kpa", "kph", "kpk", "kpl", "kpo", "kpz", "kqg", "kqh", "kqk",
    "kqm", "kqn", "kqo", "kqp", "kqs", "kqu", "kqx", "kqy", "kqz", "krh",
    "krn", "krp", "krs", "krt", "krw", "krx", "ksa", "ksb", "ksf", "ksm",
    "kso", "ksp", "ksq", "kss", "kst", "ksv", "ktb", "ktc", "ktf", "kth",
    "ktj", "kty", "ktz", "kua", "kub", "kug", "kuh", "kuj", "kul", "kun",
    "kus", "kuw", "kvf", "kvi", "kvj", "kvm", "kwb", "kwc", "kwg", "kwl",
    "kwm", "kwn", "kwp", "kws", "kwu", "kwv", "kwy", "kwz", "kxb", "kxc",
    "kxh", "kxj", "kxx", "kya", "kye", "kyf", "kym", "kyq", "kza", "kzc",
    "kzn", "kzo", "kzr", "kzy", "laf", "lag", "lai", "laj", "lak", "lal",
    "lam", "lan", "lap", "lar", "las", "lbi", "lch", "lda", "ldb", "ldd",
    "ldg", "ldh", "ldi", "ldj", "ldk", "ldl", "ldm", "ldo", "ldp", "ldq",
    "lea", "leb", "led", "lee", "lef", "leh", "lej", "lel", "lem", "leo",
    "les", "lfa", "lgg", "lgm", "lgn", "lgq", "lgz", "lia", "lie", "lig",
    "lik", "lin", "lip", "liq", "liu", "liy", "liz", "lkb", "lke", "lko",
    "lkr", "lks", "lky", "lla", "llb", "llc", "lli", "lln", "lma", "lmd",
    "lme", "lmi", "lmp", "lmx", "lna", "lnb", "lnl", "lno", "lns", "lnu",
    "lnz", "lob", "lof", "log", "loh", "loi", "lok", "lol", "lom", "lon",
    "loo", "lop", "loq", "lor", "lot", "loz", "lpx", "lri", "lrm", "lro",
    "lse", "lsm", "lth", "lto", "lts", "lua", "lub", "luc", "lue", "lug",
    "luj", "lul", "lum", "lun", "luo", "lup", "luq", "luw", "lwa", "lwg",
    "lwo", "lyn", "mae", "maf", "mas", "maw", "mbm", "mbo", "mbu", "mbv",
    "mcj", "mck", "mcn", "mcp", "mcs", "mct", "mcu", "mcw", "mcx", "mda",
    "mdd", "mde", "mdg", "mdi", "mdj", "mdk", "mdm", "mdn", "mdp", "mdq",
    "mdt", "mdu", "mdw", "mdx", "mdy", "mea", "mei", "men", "meq", "mer",
    "mes", "mev", "mew", "mey", "mfc", "mfd", "mff", "mfg", "mfh", "mfi",
    "mfj", "mfk", "mfl", "mfm", "mfn", "mfo", "mfq", "mfu", "mfv", "mfx",
    "mfz", "mgb", "mgc", "mgd", "mge", "mgg", "mgh", "mgi", "mgj", "mgn",
    "mgo", "mgq", "mgr", "mgs", "mgv", "mgw", "mgy", "mgz", "mhb", "mhd",
    "mhi", "mhk", "mhm", "mho", "mhw", "mif", "mij", "mje", "mjh", "mjs",
    "mka", "mkf", "mkk", "mkl", "mko", "mku", "mlb", "mlg", "mlj", "mlk",
    "mlo", "mlq", "mlr", "mls", "mlt", "mlw", "mma", "mmf", "mmu", "mmy",
    "mmz", "mne", "mnf", "mnh", "mnk", "mny", "moa", "moi", "moj", "mor",
    "mos", "mou", "mow", "moy", "moz", "mpa", "mpe", "mpg", "mpi", "mpk",
    "mqb", "mql", "mqu", "mrt", "mru", "msc", "mse", "msj", "msv", "msw",
    "mtb", "mtk", "mtl", "mua", "mub", "muc", "mug", "muh", "muj", "muo",
    "mur", "muu", "muy", "muz", "mvh", "mvu", "mvw", "mvz", "mwe", "mwk",
    "mwm", "mwn", "mws", "mwu", "mwz", "mxc", "mxf", "mxg", "mxh", "mxl",
    "mxo", "mxu", "mxx", "myb", "myc", "mye", "myf", "myg", "myj", "myk",
    "mym", "myo", "mys", "myx", "mzb", "mzd", "mzj", "mzk", "mzm", "mzv",
    "mzw", "naj", "naq", "nar", "nat", "naw", "nba", "nbb", "nbd", "nbh",
    "nbl", "nbm", "nbo", "nbp", "nbr", "nbv", "nbw", "ncr", "ncu", "nda",
    "ndb", "ndc", "ndd", "nde", "ndg", "ndh", "ndi", "ndj", "ndk", "ndl",
    "ndm", "ndn", "ndo", "ndp", "ndq", "ndr", "ndt", "ndu", "ndv", "ndw",
    "ndy", "ndz", "neb", "ned", "ney", "nfd", "nfr", "nfu", "nga", "ngb",
    "ngc", "ngd", "nge", "ngg", "ngh", "ngi", "ngj", "ngl", "ngn", "ngo",
    "ngp", "ngq", "ngs", "ngv", "ngw", "ngx", "ngy", "ngz", "nhb", "nhr",
    "nhu", "nie", "nih", "nim", "nin", "niq", "nix", "niy", "nja", "njd",
    "njj", "njl", "njr", "njx", "njy", "nka", "nkc", "nkn", "nko", "nkq",
    "nkt", "nku", "nkv", "nkw", "nkx", "nkz", "nla", "nle", "nlj", "nlo",
    "nlu", "nmc", "nmd", "nmg", "nmi", "nmj", "nml", "nmn", "nmq", "nmr",
    "nmz", "nnb", "nnc", "nne", "nnh", "nnj", "nnn", "nnq", "nnu", "nnw",
    "nnz", "noq", "now", "noy", "noz", "nqg", "nqk", "nql", "nqo", "nra",
    "nrb", "nsc", "nse", "nsg", "nsh", "nso", "nsx", "nte", "nti", "ntk",
    "ntm", "nto", "ntr", "nue", "nuh", "nui", "nuj", "nup", "nus", "nuu",
    "nuv", "nvo", "nwb", "nwe", "nwm", "nxd", "nxi", "nxo", "nya", "nyb",
    "nyc", "nyd", "nye", "nyf", "nyg", "nyi", "nyj", "nyk", "nym", "nyn",
    "nyo", "nyp", "nyr", "nyu", "nyy", "nza", "nzb", "nzd", "nzi", "nzk",
    "nzu", "nzy", "nzz", "obl", "obu", "oda", "odu", "ofu", "ogb", "ogc",
    "ogg", "ogo", "ogu", "okb", "okd", "oke", "oki", "okr", "oks", "oku",
    "okx", "old", "olm", "olu", "omi", "oml", "omt", "opa", "orc", "org",
    "orm", "orr", "orx", "oso", "ost", "otr", "oua", "oub", "oyd", "ozm",
    "pae", "pai", "pbi", "pbl", "pbn", "pbo", "pbp", "pbr", "pcn", "pcw",
    "pem", "pfe", "pgs", "phm", "pic", "pil", "pip", "piw", "piy", "pkb",
    "pko", "plj", "plr", "pmb", "pmm", "pmn", "pnd", "png", "pnl", "pnq",
    "pny", "pnz", "pof", "poy", "ppp", "pqa", "pug", "puu", "pwb", "pye",
    "pym", "rag", "ras", "rax", "reg", "rel", "rer", "res", "rif", "rim",
    "rin", "rkm", "rnd", "rng", "rnw", "rod", "rof", "rou", "rub", "ruc",
    "ruf", "rui", "ruk", "run", "ruy", "ruz", "rwk", "rwm", "rzh", "saa",
    "sad", "saf", "sag", "sak", "saq", "sav", "say", "sba", "sbd", "sbf",
    "sbj", "sbk", "sbm", "sbp", "sbs", "sbw", "sby", "sbz", "scv", "scw",
    "sde", "sdj", "sds", "seb", "sef", "seg", "seh", "sen", "sep", "seq",
    "ses", "sev", "sfw", "sgc", "sgi", "sgm", "sgw", "sha", "shc", "she",
    "shg", "shi", "shj", "shk", "sho", "shq", "shr", "shu", "shw", "shy",
    "shz", "sid", "sie", "sif", "sig", "sil", "sir", "siz", "sjg", "sjs",
    "skq", "skt", "sld", "slx", "smd", "smx", "sna", "snf", "sng", "snk",
    "snm", "snq", "snw", "soc", "sod", "soe", "soh", "sok", "som", "soo",
    "sop", "sor", "sos", "sot", "sox", "soy", "soz", "spp", "spy", "sqa",
    "sqh", "sqm", "srr", "ssc", "ssl", "ssn", "ssw", "ssy", "stj", "stv",
    "sub", "suj", "suk", "suq", "sur", "sus", "suw", "swa", "swb", "swc",
    "swf", "swh", "swj", "swk", "swn", "swq", "swy", "sxb", "sxe", "sxs",
    "sxw", "syc", "syi", "syk", "sym", "sys", "syx", "sze", "szg", "szv",
    "tag", "tak", "tal", "tan", "tap", "taq", "tax", "taz", "tbi", "tbm",
    "tbr", "tbt", "tbz", "tcc", "tcd", "tck", "tda", "tde", "tdk", "tdl",
    "tdo", "tdq", "tdv", "tec", "ted", "teg", "tek", "tem", "teo", "teq",
    "teu", "tex", "tey", "tez", "tfi", "tga", "tgd", "tgw", "tgy", "thk",
    "thu", "thv", "thy", "thz", "tia", "tic", "tig", "tii", "tik", "tiq",
    "tir", "tiv", "tja", "tjn", "tjo", "tke", "tkq", "tlj", "tll", "tlo",
    "tma", "tmc", "tms", "tmv", "tng", "tnr", "tny", "tod", "tog", "toh",
    "toi", "toq", "tor", "toz", "tpm", "tqq", "tqr", "trj", "tsa", "tsb",
    "tsc", "tsh", "tsn", "tso", "tsp", "tst", "tsv", "tsw", "ttb", "ttf",
    "ttj", "ttl", "ttq", "ttr", "tug", "tui", "tul", "tum", "tuq", "tuv",
    "tuy", "tuz", "tvd", "tvs", "tvu", "twi", "twl", "twn", "two", "twq",
    "twx", "txj", "tye", "tyi", "tyu", "tyx", "tzm", "uba", "ubi", "uda",
    "udl", "udu", "uha", "uiv", "uji", "ukh", "ukp", "ukq", "uku", "ukw",
    "ula", "ulb", "umb", "umm", "une", "urh", "usk", "uss", "uta", "uth",
    "utr", "uya", "vae", "vag", "vai", "vaj", "vau", "vem", "ven", "ver",
    "vid", "vif", "vig", "vin", "vit", "vki", "vkj", "vmk", "vmr", "vmw",
    "vor", "vum", "vun", "vut", "wal", "wan", "wav", "wbf", "wbh", "wbi",
    "wbj", "wci", "wdd", "wec", "weh", "wem", "wib", "wja", "wji", "wka",
    "wlc", "wle", "wll", "wlx", "wma", "wmw", "wni", "wob", "wof", "wok",
    "wol", "wom", "won", "woy", "wrn", "wss", "wti", "wud", "wum", "wun",
    "wwa", "www", "xab", "xam", "xan", "xdo", "xed", "xeg", "xel", "xho",
    "xii", "xkb", "xkg", "xkt", "xku", "xkv", "xma", "xmb", "xmc", "xmd",
    "xmg", "xmj", "xnz", "xoc", "xog", "xom", "xon", "xpe", "xrb", "xsh",
    "xsj", "xsm", "xsn", "xsq", "xtc", "xuo", "xuu", "xwe", "xwg", "xwl",
    "yaf", "yaj", "yal", "yam", "yao", "yas", "yat", "yav", "yax", "yay",
    "yaz", "yba", "ybb", "ybj", "ybl", "yei", "yel", "yer", "yes", "yey",
    "yko", "yky", "ymg", "ymk", "ymm", "yng", "ynq", "yns", "yom", "yor",
    "yot", "yre", "yul", "yun", "zag", "zah", "zaj", "zak", "zay", "zaz",
    "zdj", "zen", "zga", "zgh", "zhi", "zhw", "zil", "zim", "zin", "zir",
    "ziw", "ziz", "zmb", "zmf", "zmn", "zmo", "zmp", "zmq", "zms", "zmw",
    "zmx", "zmz", "zna", "zne", "zns", "zrn", "zua", "zul", "zuy", "zwa",
]
# fmt: on


unencoded_variants = {
    "tod_Latn": ["ʋ", "Ʋ"],
    "xpe_Latn": ["Ɓ"],
    "lom_Latn": ["Ɓ"],
    "dnj_Latn": ["Ɓ"],
    "gaa_Latn": ["Ʃ", "Ʒ"],
}

available_languages = {isoconv.get(tag, tag) + "_Latn": tag for tag in afr_tags}


class AfricanLatinProvider:
    @classmethod
    def fill(cls, language: dict) -> None:
        if language["script"] != "Latn":
            return
        if language["id"] not in available_languages:
            return

        exemplar_chars = language.get("exemplarChars", {})
        bases = exemplar_chars.get("base", "").split() or []
        auxiliary = exemplar_chars.get("auxiliary", "").split() or []

        language["shaperglot_checks"].extend(
            cls.build_no_orphaned_marks(bases, auxiliary)
        )
        language["shaperglot_checks"].extend(
            cls.build_unencoded_variants(language, bases, auxiliary)
        )
        language["shaperglot_checks"].extend(cls.build_sd_smcp(bases, auxiliary))

    @classmethod
    def build_no_orphaned_marks(
        cls, bases: List[str], auxiliary: List[str]
    ) -> Iterator[ShaperglotCheck]:
        basemarks = []
        for base in bases:
            if len(base) > 1:
                unpack = base.replace("{", "").replace("}", "")
                for char in unpack:
                    if unicodedata.combining(char):
                        basemarks.append(unpack)
        for aux in auxiliary:
            if len(aux) > 1:
                unpack = aux.replace("{", "").replace("}", "")
                for char in unpack:
                    if unicodedata.combining(char):
                        basemarks.append(unpack)
        basemark_string = "".join(basemarks)
        if not basemark_string:
            return
        yield NoOrphanedMarksCheck(
            {"check": "no_orphaned_marks", "input": {"text": basemark_string}}
        )
        yield NoOrphanedMarksCheck(
            {
                "check": "no_orphaned_marks",
                "input": {"text": basemark_string, "features": {"smcp": True}},
                "conditions": {"features": ['smcp']},
            }
        )

    @classmethod
    def build_unencoded_variants(
        cls, language: dict, bases: List[str], auxiliary: List[str]
    ) -> Iterator[ShaperglotCheck]:
        if "Ŋ" in bases or 'Ŋ' in auxiliary:
            yield UnencodedVariantsCheck(
                {"check": "unencoded_variants", "input": {"text": 'Ŋ'}}
            )
        for char in unencoded_variants.get(language["id"], []):
            yield UnencodedVariantsCheck(
                {"check": "unencoded_variants", "input": {"text": char}}
            )

    @classmethod
    def build_sd_smcp(
        cls, bases: List[str], auxiliary: List[str]
    ) -> Iterator[ShaperglotCheck]:
        for letter in bases + auxiliary:
            if len(letter) == 1 and unicodedata.category(letter) == 'Ll':
                yield ShapingDiffersCheck(
                    {
                        'check': 'shaping_differs',
                        'inputs': [
                            {'text': letter},
                            {'text': letter, 'features': {'smcp': True}},
                        ],
                        'conditions': {'features': ['smcp']},
                        'rationale': "Small-cap " + letter + " is formed",
                    }
                )



import argparse
import os
import subprocess
import glob
import yaml
import json
from multiprocessing import Pool
from pathlib import Path
from itertools import repeat
import datetime

from shaperglot.languages import Languages
from shaperglot.checker import Checker
from shaperglot.reporter import Result

gflangs = Languages()


with open("./data/iso639-3-afr-all.txt", "r") as f2:
    afr_tags = f2.read().splitlines()

results = open("results.json", "w", encoding="utf8")
overview = open("afr_tag_overview.json", "w", encoding='utf8')

isoconv = {
    "aar": "aa",
    "afr": "af",
    "aka": "ak",
    "amh": "am",
    "bam": "bm",
    "ewe": "ee",
    "ful": "ff",
    "hau": "ha",
    "her": "hz",
    "ibo": "ig",
    "kik": "ki",
    "kin": "rw",
    "kon": "kg",
    "kua": "kj",
    "lin": "ln",
    "lub": "lu",
    "lug": "lg",
    "mlg": "mg",
    "nbl": "nr",
    "nde": "nd",
    "ndo": "ng",
    "nya": "ny",
    "orm": "om",
    "run": "rn",
    "sag": "sg",
    "sna": "sn",
    "som": "so",
    "sot": "st",
    "ssw": "ss",
    "swa": "sw",
    "tir": "ti",
    "tsn": "tn",
    "tso": "ts",
    "twi": "tw",
    "ven": "ve",
    "wol": "wo",
    "xho": "xh",
    "yor": "yo",
}


def split_dict(input_dict: dict, num_parts: int) -> list:
    list_len: int = len(input_dict)
    return [
        dict(
            list(input_dict.items())[
                i * list_len // num_parts : (i + 1) * list_len // num_parts
            ]
        )
        for i in range(num_parts)
    ]


def summarize(failing_fonts, passing_fonts):
    pass_fails = []
    failkeys = set([key for l in failing_fonts for key in l.keys()])
    failfonts = [[d[x] for d in failing_fonts if x in d] for x in failkeys]

    passkeys = set([key for l in passing_fonts for key in l.keys()])
    passfonts = [[d[x] for d in passing_fonts if x in d] for x in passkeys]

    for idk, key in enumerate(passkeys):
        this_passfonts = passfonts[idk]
        if this_passfonts:
            pass_fails.append(
                {
                    'tag': key,
                    'pass': {'count': len(this_passfonts), 'fonts': this_passfonts},
                }
            )

    for idk, key in enumerate(failkeys):
        this_failfonts = failfonts[idk]
        id_in_passfails = next(
            (index for index, d in enumerate(pass_fails) if key in d['tag']), None
        )
        if this_failfonts and id_in_passfails != None:
            pass_fails[id_in_passfails]['fail'] = {
                'count': len(this_failfonts),
                'fonts': this_failfonts,
            }
        elif not this_failfonts and id_in_passfails != None:
            pass_fails[id_in_passfails]['fail'] = {'count': 0, 'fonts': []}
        elif this_failfonts and id_in_passfails == None:
            pass_fails.append(
                {
                    'tag': key,
                    'pass': {'count': 0, 'fonts': []},
                    'fail': {'count': len(this_failfonts), 'fonts': this_failfonts},
                }
            )
    return pass_fails


def check_one(font, available_tags):
    checker = Checker(font)
    print(f"analyzing {font}")
    results_for_font = {}
    for tag, item in available_tags.items():
        try:
            results_for_font[tag] = checker.check(gflangs[item])
        except:
            print(f"Something went wrong with {font}")
    return results_for_font


def run_checker(fonts, tag_results, args, failing_fonts, passing_fonts):
    with Pool() as p:
        all_font_results = p.starmap(check_one, args)
    for font, font_results in zip(fonts, all_font_results):
        fontname = Path(font).stem
        for tag, this_tag_results in font_results.items():
            fails = [
                message.message
                for message in this_tag_results
                if message.result != Result.PASS
            ]
            if fails:
                tag_results.setdefault(tag, {})[fontname] = fails
                failing_fonts.append({tag: fontname})
            else:
                passing_fonts.append({tag: fontname})


def main(args=None):
    parser = argparse.ArgumentParser(
        description="Check a library for African language support"
    )
    parser.add_argument('directory', help='directory to check')

    args = parser.parse_args()
    fonts = [str(f) for f in Path(args.directory).glob("**/*.ttf")]

    tag_results = {}
    missing_tags = []
    available_tags = {}
    passing_fonts = []
    failing_fonts = []

    for tag in afr_tags:
        if isoconv.get(tag) is not None:
            item = isoconv.get(tag) + "_Latn"
        else:
            item = tag + "_Latn"
        if item not in gflangs:
            missing_tags.append(item)
            continue

        available_tags[tag] = item

    split_avail_tags = split_dict(available_tags, 50)

    for idx, x in enumerate(split_avail_tags):
        args = zip(fonts, repeat(split_avail_tags[idx]))
        run_checker(fonts, tag_results, args, failing_fonts, passing_fonts)

    pf = summarize(failing_fonts, passing_fonts)
    pf.append({"Missing GFLang Data": missing_tags})
    current_time = datetime.datetime.now()
    pf.append({"Timestamp": f"{current_time}"})
    json.dump(tag_results, results, indent=1)
    json.dump(pf, overview, indent=1)


if __name__ == "__main__":
    main()



import pytest
from shaperglot.languages import definitions_directory, load_shaperglot_definition

tests = [ l.stem for l in definitions_directory.glob("*.yaml") ]

@pytest.mark.parametrize("language", tests)
def test_valid(language):
    load_shaperglot_definition(language, validate=True)

